/*
This file is part of Snap to Grid.

Snap to Grid is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

Snap to Grid is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
Snap to Grid. If not, see <https://www.gnu.org/licenses/>. 
*/

using HarmonyLib;
using HMLLibrary;
using RaftModLoader;
using System.Runtime.CompilerServices;
using System.Text.RegularExpressions;
using UnityEngine;
using Debug = UnityEngine.Debug;

namespace snap_to_grid
{
  public class main : Mod
  {
    static bool ExtraSettingsAPI_Loaded = false;
    public void ExtraSettingsAPI_Load() => load();
    public void ExtraSettingsAPI_SettingsClose() => load();

    // Combobox methods.
    [MethodImpl(MethodImplOptions.NoInlining)]
    static string[] ExtraSettingsAPI_GetComboboxContent(string var) =>
      new string[0];
    [MethodImpl(MethodImplOptions.NoInlining)]
    static int ExtraSettingsAPI_GetComboboxSelectedIndex(string var) => -1;
    [MethodImpl(MethodImplOptions.NoInlining)]
    public static void ExtraSettingsAPI_SetComboboxSelectedIndex
    (
      string var,
      int value
    ){ }

    // Keybinding methods.
    [MethodImpl(MethodImplOptions.NoInlining)]
    static string ExtraSettingsAPI_GetKeybindName(string var) => null;

    // Input methods.
    [MethodImpl(MethodImplOptions.NoInlining)]
    static string ExtraSettingsAPI_GetInputValue(string var) => null;

    const int CUSTOM_SCALE_IDX = 4;
    static float[] GRID_SCALES = new float[]
    {
      1.5f, // Whole
      0.75f, // Half
      0.375f, // Quarter
      0.1875f, // Eigth
      0.01f // Custom
    };
    static readonly string ID = "com.deciusmagnus.snap_to_grid";
    static readonly Sprite NOTIFY_ICON = HLib.bundle
      .LoadAsset<Sprite>("IconInfo");

    static CanvasHelper canvas;
    static int cur_scale = 2;
    static Harmony harmony;
    static bool ignore;
    static string keybind_ignore = null;
    static string keybind_switch = null;
    static HNotify notify = HNotify.Get();

    public void Start()
    {
      harmony = new Harmony(ID);
      harmony.PatchAll();
      load();
      Debug.Log("Mod snap_to_grid has been loaded!");
    }

    public void OnModUnload()
    {
      harmony.UnpatchAll(harmony.Id);
      Debug.Log("Mod snap_to_grid has been unloaded!");
    }

    public void Update()
    {
      if (!can_open_menu())
        return;

      if (MyInput.GetButtonDown(keybind_switch))
      {
        cur_scale = (cur_scale + 1) % GRID_SCALES.Length;
        ExtraSettingsAPI_SetComboboxSelectedIndex("cur_scale", cur_scale);
        string msg = $"Scale: {grid_scale_name()}";
        notify.AddNotification
        (
          HNotify.NotificationType.normal,
          msg,
          2,
          NOTIFY_ICON
        );
      }
      else if (MyInput.GetButtonDown(keybind_ignore))
        ignore = true;
      else if (MyInput.GetButtonUp(keybind_ignore))
        ignore = false;
    }

    [ConsoleCommand(name: "s2g", docs: "Snap-to-grid super command.")]
    public static void s2g(string[] args)
    {
      if (args.Length < 1)
      {
        Debug.Log("usage: s2g {command} [args ...]");
        return;
      }

      if (args[0] == "ls" || args[0] == "listscales")
      {
        string msg = "GRID_SCALES =\n[";
        for (var i = 0; i < GRID_SCALES.Length; ++i)
        {
          msg += $"\n  [{i}] = {GRID_SCALES[i]}";
        }
        msg += "\n]";
        Debug.Log(msg);
      }
      else
      {
        Debug.Log("commands: "
          + "\nhelp"
          + "\n  Shows this help."
          + "\nls, listscales"
          + "\n  List scales");
      }
    }

    static float grid_scale() => GRID_SCALES[cur_scale];
    static string grid_scale_name()
    {
      return string.Format("{0} ({1})",
        ExtraSettingsAPI_GetComboboxContent("cur_scale")[cur_scale],
        grid_scale()
      );
    }

    void load()
    {
      cur_scale = ExtraSettingsAPI_GetComboboxSelectedIndex("cur_scale");
      if (cur_scale < 0 || GRID_SCALES.Length <= cur_scale)
        cur_scale = 0;

      if (float.TryParse(ExtraSettingsAPI_GetInputValue("grid_scale"),
        out float f))
      {
        GRID_SCALES[CUSTOM_SCALE_IDX] = f;
      }
      keybind_ignore = ExtraSettingsAPI_GetKeybindName("keybind_ignore");
      keybind_switch = ExtraSettingsAPI_GetKeybindName("keybind_switch");
    }

    static bool can_open_menu()
    {
      if (canvas is null)
      {
        canvas = Traverse.Create(typeof(Helper))
            .Field<Network_Player>("localPlayer")?.Value?.canvas;
      }

      return canvas?.CanOpenMenu ?? false;
    }

    [HarmonyPatch(typeof(BlockCreator), "SetGhostBlockPositionAndRotation")]
    static class patch_BlockCreator_SetGhostBlockPositionAndRotation
    {
      static void Prefix(BlockCreator __instance)
      {
        if (ignore)
          return;

        var selected_block = __instance.selectedBlock;
        var quad_at_cursor = Traverse.Create(__instance)
          .Field<BlockQuad>("quadAtCursor").Value;
        if (selected_block is null || selected_block.snapsToQuads
          || quad_at_cursor is null || quad_at_cursor.snapToQuadPosition)
        {
          return;
        }
        
        var quad_hit = Traverse.Create(__instance)
          .Field<RaycastHit>("quadHit").Value;
        var new_pnt = quad_at_cursor.transform.parent
          .InverseTransformPoint(quad_hit.point);
        var scale = grid_scale();
        if (scale > 0)
        {
          new_pnt = new Vector3
          (
            Mathf.RoundToInt(new_pnt.x / scale),
            new_pnt.y / scale,
            Mathf.RoundToInt(new_pnt.z / scale)
          );

          quad_hit.point = quad_at_cursor.transform.parent
            .TransformPoint(new_pnt * scale);
          Traverse.Create(__instance).Field("quadHit").SetValue(quad_hit);
        }
      }
    }
  }
}